from pathlib import Path
from pandas import DataFrame as df
from datetime import date, datetime
from pandas_bokeh import *
from bokeh.io import output_file, show
from bokeh.layouts import layout, column
from bokeh.models import ColumnDataSource, DataTable, TableColumn, Panel, Tabs, NumberFormatter
from fw_ddsm.tracker import *

class Show:

    def __init__(self, output_root_folder="results/"):

        self.output_folder = ""
        self.aggregator_tracker = Tracker()
        self.aggregator_final = Tracker()
        self.community_tracker = Tracker()
        self.community_final = Tracker()

        if not output_root_folder.endswith("/"):
            output_root_folder += "/"
        self.output_root_folder = output_root_folder

        this_date = str(date.today())
        this_time = str(datetime.now().time().strftime("%H-%M-%S"))
        self.output_parent_folder = f"{self.output_root_folder}{this_date}-{this_time}/"

    def set_output_folder(self,
                          num_households=no_households,
                          inconvenience_cost_weight=care_f_weight,
                          num_dependent_tasks=no_tasks_dependent,
                          num_full_flex_task_min=no_full_flex_tasks_min,
                          num_semi_flex_task_min=no_semi_flex_tasks_min):

        self.output_folder \
            = f"{self.output_parent_folder}/h{num_households}-w{inconvenience_cost_weight}-dt{num_dependent_tasks}" \
            f"-fft{num_full_flex_task_min}-sft{num_semi_flex_task_min}/"
        path = Path(self.output_folder)
        if not path.exists():
            path.mkdir(mode=0o777, parents=True, exist_ok=False)
        return self.output_folder

    def save_data_to_files(self, algorithms, aggregator_tracker, community_tracker, aggregator_final,
                           community_final=None,
                           print_demands=False, print_prices=False, print_summary=True):
        self.aggregator_tracker = aggregator_tracker
        self.aggregator_final = aggregator_final
        self.community_tracker = community_tracker
        self.community_final = community_final

        agg_demands, agg_prices, agg_others, agg_times = aggregator_tracker.extract_data()
        agg_demands_final, agg_prices_final, agg_others_final, agg_times_final = aggregator_final.extract_data()
        community_demands, community_prices, community_others, community_times = community_tracker.extract_data()

        plot_layout = []
        plot_final_layout = []
        x_ticks = [0] + [i for i in range(1, 48) if (i + 1) % 12 == 0 ]
        # x_tick_labels = [0] + [f"{int((i + 1)/2)}h" for i in range(1, 48) if (i + 1) % 12 == 0 ]
        df_overview = df()
        for alg in algorithms.values():
            scheduling_method = alg[k2_before_fw]
            pricing_method = alg[k2_after_fw]

            # ------------------------------ FW results ------------------------------
            df_demands = df.from_dict(agg_demands[pricing_method]).div(1000)
            df_prices = df.from_dict(agg_prices[pricing_method])
            df_others = df.from_dict(agg_others[pricing_method])

            # draw graphs
            p_demands = df_demands.iloc[:, [0, df_demands.columns[-1]]] \
                .plot_bokeh(kind="line", xlabel="Time period", ylabel="Demand (kWh)",
                            title=algorithm_full_names[pricing_method],
                            plot_data_points=True,
                            xticks=x_ticks,
                            show_figure=False
                            )
            p_prices = df_prices.iloc[:, [0, df_prices.columns[-1]]] \
                .plot_bokeh(kind="line", xlabel="Time period", ylabel="Price (dollar)",
                            title=algorithm_full_names[pricing_method],
                            plot_data_points=True,
                            xticks=x_ticks,
                            show_figure=False
                            )

            # data table
            source = ColumnDataSource(df_others)
            columns = [TableColumn(field=x, title=x.replace("_", " "), formatter=NumberFormatter(format="0.00"))
                       for x in df_others.columns]
            data_table = DataTable(source=source, columns=columns)
            plot_layout.append([p_demands, p_prices, data_table])

            # ------------------------------ final schedules ------------------------------
            df_demands_final = df.from_dict(agg_demands_final[pricing_method]).div(1000)
            df_prices_final = df.from_dict(agg_prices_final[pricing_method])
            df_others_final = df.from_dict(agg_others_final[pricing_method])

            p_demands_final = df_demands_final \
                .plot_bokeh(kind="line", xlabel="Time period", ylabel="Demand (kWh)",
                            title=algorithm_full_names[pricing_method],
                            plot_data_points=True,
                            xticks=x_ticks,
                            show_figure=False
                            )
            p_prices_final = df_prices_final \
                .plot_bokeh(kind="line", xlabel="Time period", ylabel="Price (dollar)",
                            title=algorithm_full_names[pricing_method],
                            plot_data_points=True,
                            xticks=x_ticks,
                            show_figure=False
                            )

            # data table
            source_final = ColumnDataSource(df_others_final)
            columns_final = [TableColumn(field=x, title=x.replace("_", " "), formatter=NumberFormatter(format="0.00"))
                       for x in df_others_final.columns]
            data_table_final = DataTable(source=source_final, columns=columns_final)
            plot_final_layout.append([p_demands_final, p_prices_final, data_table_final])

            # ------------------------------ write all to CSV ------------------------------
            if print_demands:
                df_demands.to_csv(r"{}{}_aggregator_demands.csv".format(self.output_folder, pricing_method))
                df_demands_final.to_csv(r"{}{}_aggregator_demands_final.csv".format(self.output_folder, pricing_method))
            if print_prices:
                df_prices.to_csv(r"{}{}_aggregator_prices.csv".format(self.output_folder, pricing_method))
                df_prices_final.to_csv(r"{}{}_aggregator_prices_final.csv".format(self.output_folder, pricing_method))
            if print_summary:
                df_others.to_csv(r"{}{}_aggregator_others.csv".format(self.output_folder, pricing_method))
                df_others_final.to_csv(r"{}{}_aggregator_others_final.csv".format(self.output_folder, pricing_method))

        agg_times.update(community_times)
        df_times = df.from_dict(agg_times)
        df_times.to_csv(r"{}run_times.csv".format(self.output_folder))

        output_file(f"{self.output_folder}plots.html")
        tab1 = Panel(child=layout(plot_layout), title="FW-DDSM results")
        tab2 = Panel(child=layout(plot_final_layout), title="Actual schedules")
        # show(Tabs(tabs=[tab1, tab2]))
        show(Tabs(tabs=[tab2, tab1]))

        print("Data are written and graphs are painted. ")




from multiprocessing import Pool, cpu_count
import pickle
from time import time
from fw_ddsm.household import *
from fw_ddsm.tracker import *


class Community:

    def __init__(self, num_intervals=no_intervals, num_periods=no_periods):
        self.num_intervals = num_intervals
        self.num_periods = num_periods
        self.num_intervals_periods = int(num_intervals / num_periods)
        self.num_households = 0
        self.households = dict()
        self.tracker = Tracker()
        self.final = Tracker()
        self.scheduling_method = ""
        self.preferred_demand_profile = []

    def read(self, scheduling_method, read_from_folder="data/", inconvenience_cost_weight=None):
        read_from_folder = read_from_folder if read_from_folder.endswith("/") \
            else read_from_folder + "/"

        self.households = dict()
        self.preferred_demand_profile = None

        self.households = self.__existing_households(file_path=read_from_folder,
                                                     inconvenience_cost_weight=inconvenience_cost_weight)
        self.preferred_demand_profile = self.households.pop(k0_demand)
        self.num_households = len(self.households) - 1

        self.tracker.new(method=scheduling_method)
        self.tracker.update(num_record=0, method=scheduling_method,
                            demands=self.preferred_demand_profile, penalty=0)
        self.final.new(method=scheduling_method)
        print("0. The community is read. ")

    def new(self, file_preferred_demand_profile_path, file_demand_list_path, scheduling_method,
            num_households=no_households,
            max_demand_multiplier=maxium_demand_multiplier,
            num_tasks_dependent=no_tasks_dependent,
            full_flex_task_min=no_full_flex_tasks_min, full_flex_task_max=0,
            semi_flex_task_min=no_semi_flex_tasks_min, semi_flex_task_max=0,
            fixed_task_min=no_fixed_tasks_min, fixed_task_max=0,
            inconvenience_cost_weight=care_f_weight, max_care_factor=care_f_max,
            write_to_file_path=None):

        self.households = dict()
        self.preferred_demand_profile = None

        self.scheduling_method = scheduling_method
        self.num_households = num_households
        self.households, self.preferred_demand_profile \
            = self.__new_households(file_preferred_demand_profile_path,
                                    file_demand_list_path,
                                    max_demand_multiplier=max_demand_multiplier,
                                    num_tasks_dependent=num_tasks_dependent,
                                    full_flex_task_min=full_flex_task_min,
                                    full_flex_task_max=full_flex_task_max,
                                    semi_flex_task_min=semi_flex_task_min,
                                    semi_flex_task_max=semi_flex_task_max,
                                    fixed_task_min=fixed_task_min,
                                    fixed_task_max=fixed_task_max,
                                    inconvenience_cost_weight=inconvenience_cost_weight,
                                    max_care_factor=max_care_factor)
        self.tracker.new(method=scheduling_method)
        self.tracker.update(num_record=0, method=scheduling_method, penalty=0,
                            run_time=0)
        self.final.new(method=scheduling_method)
        self.tracker.update(num_record=0, method=scheduling_method, penalty=0, run_time=0)

        self.write_to_file("data/")
        if write_to_file_path is not None:
            self.write_to_file(write_to_file_path)

        print("0. The community is created. ")

    def write_to_file(self, write_to_file_path):

        write_to_file_path = write_to_file_path if write_to_file_path.endswith("/") \
            else write_to_file_path + "/"
        path = Path(write_to_file_path)
        if not path.exists():
            path.mkdir(mode=0o777, parents=True, exist_ok=False)

        self.households[k0_demand] = self.preferred_demand_profile
        with open(f"{write_to_file_path}{file_community_pkl}", 'wb+') as f:
            pickle.dump(self.households, f, pickle.HIGHEST_PROTOCOL)
        del self.households[k0_demand]
        f.close()

    def schedule(self, num_iteration, prices, scheduling_method, model=None, solver=None, search=None, households=None):
        print(f"{num_iteration}. Scheduling {self.num_households} households, {scheduling_method}...")

        prices = self.__convert_price(prices)
        # self.tracker.update(num_record=num_iteration - 1, method=scheduling_method, prices=prices)

        if households is None:
            households = self.households
        results = self.__schedule_multiple_processing(households=households, prices=prices,
                                                      scheduling_method=scheduling_method,
                                                      model=model, solver=solver, search=search)

        aggregate_demand_profile, weighted_total_inconvenience, time_scheduling_iteration \
            = self.__retrieve_scheduling_results(results=results, num_iteration=num_iteration)

        self.tracker.update(num_record=num_iteration, method=scheduling_method,
                            penalty=weighted_total_inconvenience, run_time=time_scheduling_iteration)

        return aggregate_demand_profile, weighted_total_inconvenience, time_scheduling_iteration

    def decide_final_schedules(self, start_probability_distribution, num_sample=0):
        final_aggregate_demand_profile = [0] * self.num_intervals
        final_total_inconvenience = 0
        total_demand = 0
        for household in self.households.values():
            chosen_demand_profile, chosen_penalty \
                = Household.finalise_household(self=Household(), household_tracker_data=household[k0_tracker],
                                               scheduling_method=self.scheduling_method,
                                               probability_distribution=start_probability_distribution)
            final_aggregate_demand_profile = [x + y for x, y in
                                              zip(chosen_demand_profile, final_aggregate_demand_profile)]
            final_total_inconvenience += chosen_penalty
            total_demand += sum(chosen_demand_profile)

        # self.final.update(num_record=num_sample, method=self.scheduling_method,
        #                   demands=final_aggregate_demand_profile, penalty=final_total_inconvenience)

        return final_aggregate_demand_profile, final_total_inconvenience

    def __convert_price(self, prices):
        num_periods = len(prices)
        num_intervals_period = int(self.num_intervals / num_periods)
        if num_periods != self.num_intervals:
            prices = [p for p in prices for _ in range(num_intervals_period)]
        else:
            prices = [p for p in prices]

        return prices

    def __new_households(self, file_probability_path, file_demand_list_path,
                         max_demand_multiplier=maxium_demand_multiplier,
                         num_tasks_dependent=no_tasks_dependent,
                         full_flex_task_min=no_full_flex_tasks_min, full_flex_task_max=0,
                         semi_flex_task_min=no_semi_flex_tasks_min, semi_flex_task_max=0,
                         fixed_task_min=no_fixed_tasks_min, fixed_task_max=0,
                         inconvenience_cost_weight=care_f_weight, max_care_factor=care_f_max):
        # ---------------------------------------------------------------------- #

        # ---------------------------------------------------------------------- #

        num_households = self.num_households
        preferred_demand_profile = genfromtxt(file_probability_path, delimiter=',', dtype="float")
        list_of_devices_power = genfromtxt(file_demand_list_path, delimiter=',', dtype="float")

        households = dict()
        num_intervals = self.num_intervals
        aggregate_demand_profile = [0] * num_intervals

        for h in range(num_households):
            tasks, household_tracker = Household.new_household(self=Household(),
                                                               num_intervals=self.num_intervals,
                                                               scheduling_method=self.scheduling_method,
                                                               preferred_demand_profile=preferred_demand_profile,
                                                               list_of_devices_power=list_of_devices_power,
                                                               max_demand_multiplier=max_demand_multiplier,
                                                               num_tasks_dependent=num_tasks_dependent,
                                                               full_flex_task_min=full_flex_task_min,
                                                               full_flex_task_max=full_flex_task_max,
                                                               semi_flex_task_min=semi_flex_task_min,
                                                               semi_flex_task_max=semi_flex_task_max,
                                                               fixed_task_min=fixed_task_min,
                                                               fixed_task_max=fixed_task_max,
                                                               inconvenience_cost_weight=inconvenience_cost_weight,
                                                               max_care_factor=max_care_factor, household_id=h)
            household_profile = tasks[k0_demand]
            aggregate_demand_profile = [x + y for x, y in zip(household_profile, aggregate_demand_profile)]

            households[h] = tasks.copy()
            households[h][k0_tracker] = household_tracker.data.copy()

        return households, aggregate_demand_profile

    def __existing_households(self, file_path, inconvenience_cost_weight=None):
        # ---------------------------------------------------------------------- #
        # ---------------------------------------------------------------------- #

        with open(f"{file_path}{file_community_pkl}", 'rb') as f:
            households = pickle.load(f)
        f.close()

        if inconvenience_cost_weight is not None:
            for household in households.values():
                household["care_factor_weight"] = inconvenience_cost_weight

        return households

    def __schedule_multiple_processing(self, households, prices, scheduling_method, model, solver, search):
        t_begin = time()
        pool = Pool(cpu_count())
        results = pool.starmap_async(Household.schedule_household,
                                     [(Household(), prices, scheduling_method, household,
                                       self.num_intervals, model, solver, search)
                                      for household in households.values()]).get()
        # parameters' order: prices, scheduling_method = self.scheduling_method, household = self.tasks, num_intervals = self.num_intervals, model = model, solver = solver, search = search
        # results = pool.starmap_async(self.schedule_single_household,
        #                              [(household, prices, scheduling_method, model, solver, search)
        #                               for household in households.values()]).get()
        pool.close()
        pool.join()
        print(f"   Finish scheduling in {round(time() - t_begin)} seconds.")
        return results

    def __retrieve_scheduling_results(self, results, num_iteration):
        aggregate_demand_profile = [0] * self.num_intervals
        total_weighted_inconvenience = 0
        time_scheduling_iteration = 0
        total_demand = 0
        for res in results:
            key = res[h_key]
            demands_household = res[k0_demand]
            weighted_penalty_household = res[k0_penalty]
            time_household = res[k0_time]
            total_demand += sum(demands_household)

            aggregate_demand_profile = [x + y for x, y in zip(demands_household, aggregate_demand_profile)]
            total_weighted_inconvenience += weighted_penalty_household
            time_scheduling_iteration += time_household

            # update each household's tracker
            self.households[key][k0_tracker] \
                = Tracker.update(self=Tracker(), num_record=num_iteration, tracker=self.households[key][k0_tracker],
                                 method=self.scheduling_method, demands=demands_household,
                                 penalty=weighted_penalty_household).copy()

        return aggregate_demand_profile, total_weighted_inconvenience, time_scheduling_iteration
